use crate::util::error::*;
use crate::util::serverfn_client::Client;
use cfg_if::cfg_if;
use leptos::prelude::*;

cfg_if! {
    if #[cfg(feature = "ssr")] {
        use crate::util::backend_state::BackendState;
        use diesel::prelude::*;
        use chrono::NaiveDate;
    }
}

/// Add an album to the database
///
/// # Arguments
///
/// * `album_title` - The name of the artist to add
/// * `release_data` - The release date of the album (Optional)
/// * `image_path` - The path to the album's image file (Optional)
///
/// # Returns
///  * `Result<(), Box<dyn Error>>` - A empty result if successful, or an error
///
#[server(endpoint = "albums/add-album", client = Client)]
pub async fn add_album(
    album_title: String,
    release_date: Option<String>,
    image_path: Option<String>,
) -> BackendResult<()> {
    use crate::models::backend::NewAlbum;
    use crate::schema::albums::{self};

    let parsed_release_date = match release_date {
        Some(date) => match NaiveDate::parse_from_str(date.trim(), "%Y-%m-%d") {
            Ok(parsed_date) => Some(parsed_date),
            Err(e) => {
                return Err(
                    InputError::InvalidInput(format!("Error parsing release date: {e}")).into(),
                );
            }
        },
        None => None,
    };

    let image_path_arg = image_path.filter(|image_path| !image_path.is_empty());

    let new_album = NewAlbum {
        title: album_title,
        release_date: parsed_release_date,
        image_path: image_path_arg,
    };

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    diesel::insert_into(albums::table)
        .values(&new_album)
        .execute(&mut db_conn)
        .context("Error inserting new album into database")?;

    Ok(())
}
