use leptos::prelude::*;

use cfg_if::cfg_if;

cfg_if! {
    if #[cfg(feature = "ssr")] {
        use leptos_axum::extract;
        use axum_login::AuthSession;
        use crate::util::auth_backend::AuthBackend;
        use crate::util::backend_state::BackendState;
    }
}

use crate::api::users::UserCredentials;
use crate::models::backend::{NewUser, User};
use crate::util::error::*;
use crate::util::serverfn_client::Client;

/// Create a new user and log them in
/// Takes in a NewUser struct, with the password in plaintext
/// Returns a Result with the error message if the user could not be created
#[server(endpoint = "signup", client = Client)]
pub async fn signup(new_user: NewUser) -> BackendResult<()> {
    // Check LIBRETUNES_DISABLE_SIGNUP env var
    if std::env::var("LIBRETUNES_DISABLE_SIGNUP").is_ok_and(|v| v == "true") {
        return Err(AuthError::SignupDisabled.into());
    }

    use crate::api::users::create_user;

    // Ensure the user has no id, and is not a self-proclaimed admin
    let new_user = NewUser {
        admin: false,
        ..new_user
    };

    create_user(&new_user)
        .await
        .context("Error creating user")?;

    let mut auth_session = extract::<AuthSession<AuthBackend>>()
        .await
        .context("Error extracting auth session")?;

    let credentials = UserCredentials {
        username_or_email: new_user.username.clone(),
        password: new_user.password.clone().unwrap(),
    };

    match auth_session.authenticate(credentials).await {
        Ok(Some(user)) => auth_session
            .login(&user)
            .await
            .map_err(|e| AuthError::AuthError(format!("Error logging in user: {e}")).into()),
        Ok(None) => Err(AuthError::InvalidCredentials.into()),
        Err(e) => Err(AuthError::AuthError(format!("Error authenticating user: {e}")).into()),
    }
}

/// Log a user in
/// Takes in a username or email and a password in plaintext
/// Returns a Result with a boolean indicating if the login was successful
#[server(endpoint = "login", client = Client)]
pub async fn login(credentials: UserCredentials) -> BackendResult<Option<User>> {
    use crate::api::users::validate_user;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let mut auth_session = extract::<AuthSession<AuthBackend>>()
        .await
        .context("Error extracting auth session")?;

    let user = validate_user(credentials, &mut db_conn)
        .await
        .context("Error validating user credentials")?;

    if let Some(mut user) = user {
        auth_session
            .login(&user)
            .await
            .map_err(|e| AuthError::AuthError(format!("Error logging in user: {e}")))?;

        user.password = None;
        Ok(Some(user))
    } else {
        Ok(None)
    }
}

/// Log a user out
/// Returns a Result with the error message if the user could not be logged out
#[server(endpoint = "logout", client = Client)]
pub async fn logout() -> BackendResult<()> {
    let mut auth_session = extract::<AuthSession<AuthBackend>>()
        .await
        .context("Error extracting auth session")?;

    auth_session
        .logout()
        .await
        .map_err(|e| AuthError::AuthError(format!("Error logging out user: {e}")))?;

    leptos_axum::redirect("/login");
    Ok(())
}

/// Check if a user is logged in
/// Returns a Result with a boolean indicating if the user is logged in
#[server(endpoint = "check_auth", client = Client)]
pub async fn check_auth() -> BackendResult<bool> {
    let auth_session = extract::<AuthSession<AuthBackend>>()
        .await
        .context("Error extracting auth session")?;

    Ok(auth_session.user.is_some())
}

/// Require that a user is logged in
/// Returns a Result with the error message if the user is not logged in
/// Intended to be used at the start of a protected route, to ensure the user is logged in:
/// ```rust
/// use leptos::prelude::*;
/// use libretunes::api::auth::require_auth;
/// use libretunes::util::error::*;
/// #[server(endpoint = "protected_route")]
/// pub async fn protected_route() -> BackendResult<()> {
///     require_auth().await?;
///     // Continue with protected route
///     Ok(())
/// }
/// ```
#[cfg(feature = "ssr")]
pub async fn require_auth() -> BackendResult<()> {
    check_auth()
        .await
        .context("Error checking authentication")
        .and_then(|logged_in| {
            if logged_in {
                Ok(())
            } else {
                Err(AuthError::Unauthorized.into())
            }
        })
}

/// Get the current logged-in user
/// Returns a Result with the user if they are logged in
/// Returns an error if the user is not logged in, or if there is an error getting the user
/// Intended to be used in a route to get the current user:
/// ```rust
/// use leptos::prelude::*;
/// use libretunes::api::auth::get_user;
/// use libretunes::util::error::*;
/// #[server(endpoint = "user_route")]
/// pub async fn user_route() -> BackendResult<()> {
///     let user = get_user().await?;
///     println!("Logged in as: {}", user.username);
///     // Do something with the user
///     Ok(())
/// }
/// ```
#[cfg(feature = "ssr")]
pub async fn get_user() -> BackendResult<User> {
    let auth_session = extract::<AuthSession<AuthBackend>>()
        .await
        .context("Error extracting auth session")?;

    auth_session.user.ok_or(AuthError::Unauthorized.into())
}

#[server(endpoint = "get_logged_in_user", client = Client)]
pub async fn get_logged_in_user() -> BackendResult<Option<User>> {
    let auth_session = extract::<AuthSession<AuthBackend>>()
        .await
        .context("Error extracting auth session")?;

    let user = auth_session.user.map(|mut user| {
        user.password = None;
        user
    });

    Ok(user)
}

/// Check if a user is an admin
/// Returns a Result with a boolean indicating if the user is logged in and an admin
#[server(endpoint = "check_admin", client = Client)]
pub async fn check_admin() -> BackendResult<bool> {
    let auth_session = extract::<AuthSession<AuthBackend>>()
        .await
        .context("Error extracting auth session")?;

    Ok(auth_session.user.as_ref().map(|u| u.admin).unwrap_or(false))
}

/// Require that a user is logged in and an admin
/// Returns a Result with the error message if the user is not logged in or is not an admin
/// Intended to be used at the start of a protected route, to ensure the user is logged in and an admin:
/// ```rust
/// use leptos::prelude::*;
/// use libretunes::api::auth::require_admin;
/// use libretunes::util::error::*;
/// #[server(endpoint = "protected_admin_route")]
/// pub async fn protected_admin_route() -> BackendResult<()> {
///     require_admin().await?;
///     // Continue with protected route
///     Ok(())
/// }
/// ```
#[cfg(feature = "ssr")]
pub async fn require_admin() -> BackendResult<()> {
    check_admin().await.and_then(|is_admin| {
        if is_admin {
            Ok(())
        } else {
            Err(AuthError::AdminRequired.into())
        }
    })
}
