use crate::models::frontend;
use crate::util::error::*;
use crate::util::serverfn_client::Client;
use leptos::prelude::*;

use cfg_if::cfg_if;

cfg_if! {
if #[cfg(feature = "ssr")] {
    use diesel::sql_types::*;
    use diesel::*;
    use diesel::pg::Pg;
    use diesel::expression::AsExpression;
    use std::collections::HashMap;
    use crate::models::backend;
    use crate::util::backend_state::BackendState;

    // Define pg_trgm operators
    // Functions do not use indices for queries, so we need to use operators
    diesel::infix_operator!(Similarity, " % ", backend: Pg);
    diesel::infix_operator!(Distance, " <-> ", Float, backend: Pg);

    // Create functions to make use of the operators in queries
    fn trgm_similar<T: AsExpression<Text>, U: AsExpression<Text>>(left: T, right: U)
        -> Similarity<T::Expression, U::Expression> {
        Similarity::new(left.as_expression(), right.as_expression())
    }

    fn trgm_distance<T: AsExpression<Text>, U: AsExpression<Text>>(left: T, right: U)
        -> Distance<T::Expression, U::Expression> {
        Distance::new(left.as_expression(), right.as_expression())
    }
}
}

/// A simple type for search results
/// A vector of tuples containing the item and its score
pub type SearchResults<T> = Vec<(T, f32)>;

/// Search for albums by title
///
/// # Arguments
/// `query` - The search query. This will be used to perform a fuzzy search on the album titles
/// `limit` - The maximum number of results to return
///
/// # Returns
/// A Result containing a vector of albums if the search was successful, or an error if the search failed
#[server(endpoint = "search_albums", client = Client)]
pub async fn search_albums(
    query: String,
    limit: i64,
) -> BackendResult<SearchResults<frontend::Album>> {
    use crate::schema::*;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let album_ids = albums::table
        .filter(trgm_similar(albums::title, query.clone()))
        .order_by(trgm_distance(albums::title, query.clone()).desc())
        .limit(limit)
        .select(albums::id)
        .load::<i32>(&mut db_conn)
        .context("Error loading album ids from database")?;

    let mut albums_map: HashMap<i32, (frontend::Album, f32)> = HashMap::new();

    let album_artists: Vec<(backend::Album, backend::Artist, f32)> = albums::table
        .filter(albums::id.eq_any(album_ids))
        .inner_join(
            album_artists::table
                .inner_join(artists::table)
                .on(albums::id.eq(album_artists::album_id)),
        )
        .select((
            albums::all_columns,
            artists::all_columns,
            trgm_distance(albums::title, query.clone()),
        ))
        .load(&mut db_conn)
        .context("Error loading album artists from database")?;

    for (album, artist, score) in album_artists {
        if let Some((stored_album, _score)) = albums_map.get_mut(&album.id) {
            stored_album.artists.push(artist);
        } else {
            let albumdata = frontend::Album {
                id: album.id,
                title: album.title,
                artists: vec![artist],
                release_date: album.release_date,
                image_path: album
                    .image_path
                    .unwrap_or("/assets/images/placeholders/MusicPlaceholder.svg".to_string()),
            };

            albums_map.insert(album.id, (albumdata, score));
        }
    }

    let mut albums: Vec<(frontend::Album, f32)> = albums_map.into_values().collect();
    albums.sort_by(|(_a, a_score), (_b, b_score)| b_score.total_cmp(a_score));

    Ok(albums)
}

/// Search for artists by name
///
/// # Arguments
/// `query` - The search query. This will be used to perform a fuzzy search on the artist names
/// `limit` - The maximum number of results to return
///
/// # Returns
/// A Result containing a vector of artists if the search was successful, or an error if the search failed
#[server(endpoint = "search_artists", client = Client)]
pub async fn search_artists(
    query: String,
    limit: i64,
) -> BackendResult<SearchResults<frontend::Artist>> {
    use crate::schema::*;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let artist_list = artists::table
        .filter(trgm_similar(artists::name, query.clone()))
        .order_by(trgm_distance(artists::name, query.clone()).desc())
        .limit(limit)
        .select((artists::all_columns, trgm_distance(artists::name, query)))
        .load::<(backend::Artist, f32)>(&mut db_conn)
        .context("Error loading artists from database")?;

    let artist_data = artist_list
        .into_iter()
        .map(|(artist, score)| {
            (
                frontend::Artist {
                    id: artist.id,
                    name: artist.name,
                    image_path: format!("/assets/images/artist/{}.webp", artist.id),
                },
                score,
            )
        })
        .collect();

    Ok(artist_data)
}

/// Search for songs by title
///
/// # Arguments
/// `query` - The search query. This will be used to perform a fuzzy search on the song titles
/// `limit` - The maximum number of results to return
///
/// # Returns
/// A Result containing a vector of songs if the search was successful, or an error if the search failed
#[server(endpoint = "search_songs", client = Client)]
pub async fn search_songs(
    query: String,
    limit: i64,
) -> BackendResult<SearchResults<frontend::Song>> {
    use crate::api::auth::get_logged_in_user;
    use crate::schema::*;

    let user = get_logged_in_user()
        .await
        .context("Error getting logged-in user")?;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let song_list = if let Some(user) = user {
        let song_list: Vec<(
            backend::Song,
            Option<backend::Album>,
            Option<backend::Artist>,
            Option<(i32, i32)>,
            Option<(i32, i32)>,
            f32,
        )> = songs::table
            .filter(trgm_similar(songs::title, query.clone()))
            .order_by(trgm_distance(songs::title, query.clone()).desc())
            .limit(limit)
            .left_join(albums::table.on(songs::album_id.eq(albums::id.nullable())))
            .left_join(
                song_artists::table
                    .inner_join(artists::table)
                    .on(songs::id.eq(song_artists::song_id)),
            )
            .left_join(
                song_likes::table.on(songs::id
                    .eq(song_likes::song_id)
                    .and(song_likes::user_id.eq(user.id))),
            )
            .left_join(
                song_dislikes::table.on(songs::id
                    .eq(song_dislikes::song_id)
                    .and(song_dislikes::user_id.eq(user.id))),
            )
            .select((
                songs::all_columns,
                albums::all_columns.nullable(),
                artists::all_columns.nullable(),
                song_likes::all_columns.nullable(),
                song_dislikes::all_columns.nullable(),
                trgm_distance(songs::title, query.clone()),
            ))
            .load(&mut db_conn)
            .context("Error loading songs from database")?;

        song_list
    } else {
        let song_list: Vec<(
            backend::Song,
            Option<backend::Album>,
            Option<backend::Artist>,
            f32,
        )> = songs::table
            .filter(trgm_similar(songs::title, query.clone()))
            .order_by(trgm_distance(songs::title, query.clone()).desc())
            .limit(limit)
            .left_join(albums::table.on(songs::album_id.eq(albums::id.nullable())))
            .left_join(
                song_artists::table
                    .inner_join(artists::table)
                    .on(songs::id.eq(song_artists::song_id)),
            )
            .select((
                songs::all_columns,
                albums::all_columns.nullable(),
                artists::all_columns.nullable(),
                trgm_distance(songs::title, query.clone()),
            ))
            .load(&mut db_conn)
            .context("Error loading songs from database")?;

        song_list
            .into_iter()
            .map(|(song, album, artist, score)| (song, album, artist, None, None, score))
            .collect()
    };

    let mut search_songs: HashMap<i32, (frontend::Song, f32)> =
        HashMap::with_capacity(song_list.len());

    for (song, album, artist, like, dislike, score) in song_list {
        if let Some((stored_songdata, _score)) = search_songs.get_mut(&song.id) {
            // If the song is already in the map, update the artists
            if let Some(artist) = artist {
                stored_songdata.artists.push(artist);
            }
        } else {
            let like_dislike = match (like, dislike) {
                (Some(_), Some(_)) => Some((true, true)),
                (Some(_), None) => Some((true, false)),
                (None, Some(_)) => Some((false, true)),
                _ => None,
            };

            let image_path = song.image_path.unwrap_or(
                album
                    .as_ref()
                    .and_then(|album| album.image_path.clone())
                    .unwrap_or("/assets/images/placeholders/MusicPlaceholder.svg".to_string()),
            );

            let songdata = frontend::Song {
                id: song.id,
                title: song.title,
                artists: artist.map(|artist| vec![artist]).unwrap_or_default(),
                album,
                track: song.track,
                duration: song.duration,
                release_date: song.release_date,
                song_path: song.storage_path,
                image_path,
                like_dislike,
                added_date: song.added_date,
            };

            search_songs.insert(song.id, (songdata, score));
        }
    }

    // Sort the songs by date
    let mut songs: Vec<(frontend::Song, f32)> = search_songs.into_values().collect();
    songs.sort_by(|(_a, a_score), (_b, b_score)| b_score.total_cmp(a_score));

    Ok(songs)
}

/// Search for songs, albums, and artists by title or name
///
/// # Arguments
/// `query` - The search query. This will be used to perform a fuzzy search on the
/// song titles, album titles, and artist names
/// `limit` - The maximum number of results to return for each type
///
/// # Returns
/// A Result containing a tuple of vectors of albums, artists, and songs if the search was successful,
#[server(endpoint = "search", client = Client)]
pub async fn search(
    query: String,
    limit: i64,
) -> BackendResult<(
    SearchResults<frontend::Album>,
    SearchResults<frontend::Artist>,
    SearchResults<frontend::Song>,
)> {
    let albums = search_albums(query.clone(), limit);
    let artists = search_artists(query.clone(), limit);
    let songs = search_songs(query, limit);

    use tokio::join;

    let (albums, artists, songs) = join!(albums, artists, songs);

    let albums = albums.context("Error searching for albums")?;
    let artists = artists.context("Error searching for artists")?;
    let songs = songs.context("Error searching for songs")?;

    Ok((albums, artists, songs))
}
