use leptos::prelude::*;

use cfg_if::cfg_if;

use crate::models::frontend;
use crate::util::error::*;
use crate::util::serverfn_client::Client;

cfg_if! {
    if #[cfg(feature = "ssr")] {
        use crate::util::backend_state::BackendState;
        use crate::api::auth::get_user;
        use crate::models::backend::{Song, Album, Artist};
        use diesel::prelude::*;
    }
}

/// Like or unlike a song
#[server(endpoint = "songs/set_like", client = Client)]
pub async fn set_like_song(song_id: i32, like: bool) -> BackendResult<()> {
    let user = get_user().await.context("Error getting logged-in user")?;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    user.set_like_song(song_id, like, &mut db_conn)
        .await
        .context("Error setting like status for song")
}

/// Dislike or remove dislike from a song
#[server(endpoint = "songs/set_dislike", client = Client)]
pub async fn set_dislike_song(song_id: i32, dislike: bool) -> BackendResult<()> {
    let user = get_user().await.context("Error getting logged-in user")?;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    user.set_dislike_song(song_id, dislike, &mut db_conn)
        .await
        .context("Error setting dislike status for song")
}

/// Get the like and dislike status of a song
#[server(endpoint = "songs/get_like_dislike", client = Client)]
pub async fn get_like_dislike_song(song_id: i32) -> BackendResult<(bool, bool)> {
    let user = get_user().await.context("Error getting logged-in user")?;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    // TODO this could probably be done more efficiently with a tokio::try_join, but
    // doing so is much more complicated than it would initially seem

    let like = user
        .get_like_song(song_id, &mut db_conn)
        .await
        .context("Error getting song like status")?;

    let dislike = user
        .get_dislike_song(song_id, &mut db_conn)
        .await
        .context("Error getting song dislike status")?;

    Ok((like, dislike))
}

#[server(endpoint = "songs/get", client = Client)]
pub async fn get_song_by_id(song_id: i32) -> BackendResult<Option<frontend::Song>> {
    use crate::schema::*;

    let user_id: i32 = get_user().await.context("Error getting logged-in user")?.id;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let song_parts: Vec<(
        Song,
        Option<Album>,
        Option<Artist>,
        Option<(i32, i32)>,
        Option<(i32, i32)>,
    )> = songs::table
        .find(song_id)
        .left_join(albums::table.on(songs::album_id.eq(albums::id.nullable())))
        .left_join(
            song_artists::table
                .inner_join(artists::table)
                .on(songs::id.eq(song_artists::song_id)),
        )
        .left_join(
            song_likes::table.on(songs::id
                .eq(song_likes::song_id)
                .and(song_likes::user_id.eq(user_id))),
        )
        .left_join(
            song_dislikes::table.on(songs::id
                .eq(song_dislikes::song_id)
                .and(song_dislikes::user_id.eq(user_id))),
        )
        .select((
            songs::all_columns,
            albums::all_columns.nullable(),
            artists::all_columns.nullable(),
            song_likes::all_columns.nullable(),
            song_dislikes::all_columns.nullable(),
        ))
        .load(&mut db_conn)
        .context("Error loading song from database")?;

    let song = song_parts.first().cloned();
    let artists = song_parts
        .into_iter()
        .filter_map(|(_, _, artist, _, _)| artist)
        .collect::<Vec<_>>();

    match song {
        Some((song, album, _artist, like, dislike)) => {
            // Use song image path, or fall back to album image path, or fall back to placeholder
            let image_path = song.image_path.clone().unwrap_or_else(|| {
                album
                    .as_ref()
                    .and_then(|album| album.image_path.clone())
                    .unwrap_or("/assets/images/placeholders/MusicPlaceholder.svg".to_string())
            });

            Ok(Some(frontend::Song {
                id: song.id,
                title: song.title.clone(),
                artists,
                album: album.clone(),
                track: song.track,
                duration: song.duration,
                release_date: song.release_date,
                song_path: song.storage_path.clone(),
                image_path,
                like_dislike: Some((like.is_some(), dislike.is_some())),
                added_date: song.added_date,
            }))
        }
        None => Ok(None),
    }
}

#[server(endpoint = "songs/plays", client = Client)]
pub async fn get_song_plays(song_id: i32) -> BackendResult<i64> {
    use crate::schema::*;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let plays = song_history::table
        .filter(song_history::song_id.eq(song_id))
        .count()
        .get_result::<i64>(&mut db_conn)
        .context("Error getting song plays")?;

    Ok(plays)
}

#[server(endpoint = "songs/my-plays", client = Client)]
pub async fn get_my_song_plays(song_id: i32) -> BackendResult<i64> {
    use crate::schema::*;

    let user_id: i32 = get_user().await.context("Error getting logged-in user")?.id;

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let plays = song_history::table
        .filter(
            song_history::song_id
                .eq(song_id)
                .and(song_history::user_id.eq(user_id)),
        )
        .count()
        .get_result::<i64>(&mut db_conn)
        .context("Error getting song plays for user")?;

    Ok(plays)
}
