use crate::util::error::*;
use leptos::prelude::*;
use server_fn::codec::{MultipartData, MultipartFormData};

use cfg_if::cfg_if;

cfg_if! {
    if #[cfg(feature = "ssr")] {
        use multer::Field;
        use crate::util::backend_state::BackendState;
        use crate::util::extract_field::extract_field;
        use diesel::prelude::*;
        use log::*;
        use chrono::NaiveDate;
    }
}

/// Validate the artist ids in a multipart field
/// Expects a field with a comma-separated list of artist ids, and ensures each is a valid artist id in the database
#[cfg(feature = "ssr")]
async fn validate_artist_ids(artist_ids: Field<'static>) -> BackendResult<Vec<i32>> {
    use crate::models::backend::Artist;
    use diesel::result::Error::NotFound;

    // Extract the artist id from the field
    match artist_ids.text().await {
        Ok(artist_ids) => {
            let artist_ids = artist_ids.trim_end_matches(',').split(',');

            let mut db_conn = BackendState::get().await?.get_db_conn()?;

            artist_ids
                .filter(|artist_id| !artist_id.is_empty())
                .map(|artist_id| {
                    // Parse the artist id as an integer
                    if let Ok(artist_id) = artist_id.parse::<i32>() {
                        // Check if the artist exists
                        let artist = crate::schema::artists::dsl::artists
                            .find(artist_id)
                            .first::<Artist>(&mut db_conn);

                        match artist {
                            Ok(_) => Ok(artist_id),
                            Err(NotFound) => Err(AccessError::NotFound.context("Artist not found")),
                            Err(e) => Err(e.context("Error finding artist id")),
                        }
                    } else {
                        Err(InputError::InvalidInput("Error parsing artist id".to_string()).into())
                    }
                })
                .collect()
        }

        Err(e) => Err(InputError::FieldReadError(format!("Error reading artist ids: {e}")).into()),
    }
}

/// Validate the album id in a multipart field
/// Expects a field with an album id, and ensures it is a valid album id in the database
#[cfg(feature = "ssr")]
async fn validate_album_id(album_id: Field<'static>) -> BackendResult<Option<i32>> {
    use crate::models::backend::Album;
    use diesel::result::Error::NotFound;

    // Extract the album id from the field
    match album_id.text().await {
        Ok(album_id) => {
            if album_id.is_empty() {
                return Ok(None);
            }

            // Parse the album id as an integer
            if let Ok(album_id) = album_id.parse::<i32>() {
                let mut db_conn = BackendState::get().await?.get_db_conn()?;

                // Check if the album exists
                let album = crate::schema::albums::dsl::albums
                    .find(album_id)
                    .first::<Album>(&mut db_conn);

                match album {
                    Ok(_) => Ok(Some(album_id)),
                    Err(NotFound) => Err(AccessError::NotFound.context("Album not found")),
                    Err(e) => Err(e.context("Error finding album id")),
                }
            } else {
                Err(InputError::InvalidInput("Error parsing album id".to_string()).into())
            }
        }
        Err(e) => Err(InputError::FieldReadError(format!("Error reading album id: {e}")).into()),
    }
}

/// Validate the track number in a multipart field
/// Expects a field with a track number, and ensures it is a valid track number (non-negative integer)
#[cfg(feature = "ssr")]
async fn validate_track_number(track_number: Field<'static>) -> BackendResult<Option<i32>> {
    match track_number.text().await {
        Ok(track_number) => {
            if track_number.is_empty() {
                return Ok(None);
            }

            if let Ok(track_number) = track_number.parse::<i32>() {
                if track_number < 0 {
                    Err(
                        InputError::InvalidInput("Track number must be positive or 0".to_string())
                            .into(),
                    )
                } else {
                    Ok(Some(track_number))
                }
            } else {
                Err(InputError::InvalidInput("Error parsing track number".to_string()).into())
            }
        }
        Err(e) => {
            Err(InputError::FieldReadError(format!("Error reading track number: {e}")).into())
        }
    }
}

/// Validate the release date in a multipart field
/// Expects a field with a release date, and ensures it is a valid date in the format [year]-[month]-[day]
#[cfg(feature = "ssr")]
async fn validate_release_date(release_date: Field<'static>) -> BackendResult<Option<NaiveDate>> {
    match release_date.text().await {
        Ok(release_date) => {
            if release_date.trim().is_empty() {
                return Ok(None);
            }

            let release_date = NaiveDate::parse_from_str(release_date.trim(), "%Y-%m-%d");

            match release_date {
                Ok(release_date) => Ok(Some(release_date)),
                Err(_) => Err(InputError::InvalidInput(
                    "Invalid release date format, expected YYYY-MM-DD".to_string(),
                )
                .into()),
            }
        }
        Err(e) => Err(InputError::InvalidInput(format!("Error reading release date: {e}")).into()),
    }
}

/// Handle the file upload form
#[server(input = MultipartFormData, endpoint = "/upload")]
pub async fn upload(data: MultipartData) -> BackendResult<()> {
    // Safe to unwrap - "On the server side, this always returns Some(_). On the client side, always returns None."
    let mut data = data.into_inner().unwrap();

    let mut title = None;
    let mut artist_ids = None;
    let mut album_id = None;
    let mut track = None;
    let mut release_date = None;
    let mut file_name = None;
    let mut duration = None;

    // Fetch the fields from the form data
    while let Ok(Some(mut field)) = data.next_field().await {
        let name = field.name().unwrap_or_default().to_string();

        match name.as_str() {
            "title" => {
                title = Some(
                    extract_field(field)
                        .await
                        .context("Error extracting title field")?,
                );
            }
            "artist_ids" => {
                artist_ids = Some(
                    validate_artist_ids(field)
                        .await
                        .context("Error validating artist ids")?,
                );
            }
            "album_id" => {
                album_id = Some(
                    validate_album_id(field)
                        .await
                        .context("Error validating album id")?,
                );
            }
            "track_number" => {
                track = Some(
                    validate_track_number(field)
                        .await
                        .context("Error validating track number")?,
                );
            }
            "release_date" => {
                release_date = Some(
                    validate_release_date(field)
                        .await
                        .context("Error validating release date")?,
                );
            }
            "file" => {
                use crate::util::audio::extract_metadata;
                use std::fs::OpenOptions;
                use std::io::{Seek, Write};
                use symphonia::core::codecs::CODEC_TYPE_MP3;

                // Some logging is done here where there is high potential for bugs / failures,
                // or behavior that we may wish to change in the future

                // Create file name
                let title = title.clone().ok_or(InputError::InvalidInput(
                    "Title field must be present and must precede file field".to_string(),
                ))?;

                let clean_title = title.replace(" ", "_").replace("/", "_");
                let date_str = chrono::Utc::now().format("%Y-%m-%d_%H:%M:%S").to_string();
                let upload_path = format!("assets/audio/upload-{date_str}_{clean_title}.mp3");
                file_name = Some(format!("upload-{date_str}_{clean_title}.mp3"));

                debug!("Saving uploaded file {}", upload_path);

                // Save file to disk
                // Use these open options to create the file, write to it, then read from it
                let mut file = OpenOptions::new()
                    .read(true)
                    .write(true)
                    .create(true)
                    .truncate(true)
                    .open(upload_path.clone())
                    .context("Error opening file for upload")?;

                while let Some(chunk) = field.chunk().await.map_err(|e| {
                    InputError::FieldReadError(format!("Error reading file chunk: {e}"))
                })? {
                    file.write_all(&chunk)
                        .context("Error writing field chunk to file")?;
                }

                file.flush().context("Error flusing file")?;

                // Rewind the file so the duration can be measured
                file.rewind().context("Error rewinding file")?;

                // Get the codec and duration of the file
                let (file_codec, file_duration) = extract_metadata(file)
                    .context("Error extracting metadata from uploaded file")?;

                if file_codec != CODEC_TYPE_MP3 {
                    return Err(InputError::InvalidInput(format!(
                        "Invalid uploaded audio file codec: {file_codec}"
                    ))
                    .into());
                }

                duration = Some(file_duration);
            }
            _ => {
                warn!("Unknown file upload field: {}", name);
            }
        }
    }

    // Unwrap mandatory fields
    let title = title.ok_or(InputError::MissingField("title".to_string()))?;
    let artist_ids = artist_ids.unwrap_or(vec![]);
    let file_name = file_name.ok_or(InputError::MissingField("file".to_string()))?;
    let duration = duration.ok_or(InputError::MissingField("duration".to_string()))?;
    let duration = i32::try_from(duration)
        .map_err(|e| InputError::InvalidInput(format!("Error parsing duration: {e}")))
        .context("Error converting duration to i32")?;

    let album_id = album_id.unwrap_or(None);
    let track = track.unwrap_or(None);
    let release_date = release_date.unwrap_or(None);

    if album_id.is_some() != track.is_some() {
        return Err(InputError::InvalidInput(
            "Album id and track number must both be present or both be absent".to_string(),
        )
        .into());
    }

    // Create the song
    use crate::models::backend::{NewSong, Song};
    let song = NewSong {
        title,
        album_id,
        track,
        duration,
        release_date,
        storage_path: file_name,
        image_path: None,
    };

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    // Save the song to the database
    let song = song
        .insert_into(crate::schema::songs::table)
        .get_result::<Song>(&mut db_conn)
        .context("Error adding song to database")?;

    // Save the song's artists to the database
    use crate::schema::song_artists;
    use diesel::ExpressionMethods;

    let artist_ids = artist_ids
        .into_iter()
        .map(|artist_id| {
            (
                song_artists::song_id.eq(song.id),
                song_artists::artist_id.eq(artist_id),
            )
        })
        .collect::<Vec<_>>();

    diesel::insert_into(crate::schema::song_artists::table)
        .values(&artist_ids)
        .execute(&mut db_conn)
        .context("Error saving song artists to database")?;

    Ok(())
}
