cfg_if::cfg_if! {
    if #[cfg(feature = "ssr")] {
        use diesel::prelude::*;

        use pbkdf2::{
            password_hash::{
                rand_core::OsRng,
                PasswordHasher, PasswordHash, SaltString, PasswordVerifier, Error
            },
            Pbkdf2
        };

        use crate::models::backend::NewUser;
        use crate::util::backend_state::BackendState;
        use crate::util::database::PgPooledConn;
    }
}

use crate::models::backend::User;
use crate::util::error::*;
use crate::util::serverfn_client::Client;
use leptos::prelude::*;
use serde::{Deserialize, Serialize};

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct UserCredentials {
    pub username_or_email: String,
    pub password: String,
}

/// Get a user from the database by username or email
/// Returns a Result with the user if found, None if not found, or an error if there was a problem
#[cfg(feature = "ssr")]
pub async fn find_user(
    username_or_email: String,
    db_conn: &mut PgPooledConn,
) -> BackendResult<Option<User>> {
    use crate::schema::users::dsl::*;

    // Look for either a username or email that matches the input, and return an option with None if no user is found
    let user = users
        .filter(username.eq(username_or_email.clone()))
        .or_filter(email.eq(username_or_email))
        .first::<User>(db_conn)
        .optional()
        .context("Error loading user from database")?;

    Ok(user)
}

/// Get a user from the database by ID
/// Returns a Result with the user if found, None if not found, or an error if there was a problem
#[cfg(feature = "ssr")]
pub async fn find_user_by_id(
    user_id: i32,
    db_conn: &mut PgPooledConn,
) -> BackendResult<Option<User>> {
    use crate::schema::users::dsl::*;

    let user = users
        .filter(id.eq(user_id))
        .first::<User>(db_conn)
        .optional()
        .context("Error loading user from database")?;

    Ok(user)
}

/// Create a new user in the database
/// Returns an empty Result if successful, or an error if there was a problem
#[cfg(feature = "ssr")]
pub async fn create_user(new_user: &NewUser) -> BackendResult<()> {
    use crate::schema::users::dsl::*;

    let new_password =
        new_user
            .password
            .clone()
            .ok_or(BackendError::InputError(InputError::MissingField(
                "password".to_string(),
            )))?;

    let salt = SaltString::generate(&mut OsRng);
    let password_hash = Pbkdf2
        .hash_password(new_password.as_bytes(), &salt)
        .map_err(|e| AuthError::AuthError(format!("Error hashing password: {e}")))?
        .to_string();

    let new_user = NewUser {
        password: Some(password_hash),
        ..new_user.clone()
    };

    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    diesel::insert_into(users)
        .values(&new_user)
        .execute(&mut db_conn)
        .context("Error inserting new user into database")?;

    Ok(())
}

/// Validate a user's credentials
/// Returns a Result with the user if the credentials are valid, None if not valid, or an error if there was a problem
#[cfg(feature = "ssr")]
pub async fn validate_user(
    credentials: UserCredentials,
    db_conn: &mut PgPooledConn,
) -> BackendResult<Option<User>> {
    let db_user = find_user(credentials.username_or_email.clone(), db_conn)
        .await
        .context("Error finding user in database")?;

    // If the user is not found, return None
    let db_user = match db_user {
        Some(user) => user,
        None => return Ok(None),
    };

    let db_password = db_user.password.clone().ok_or(AuthError::AuthError(
        "No password stored for user".to_string(),
    ))?;

    let password_hash = PasswordHash::new(&db_password)
        .map_err(|e| AuthError::AuthError(format!("{e}")))
        .context("Error parsing password hash from database")?;

    match Pbkdf2.verify_password(credentials.password.as_bytes(), &password_hash) {
        Ok(()) => {}
        Err(Error::Password) => {
            return Ok(None);
        }
        Err(e) => {
            return Err(
                AuthError::AuthError(format!("{e}")).context("Error verifying password hash")
            );
        }
    }

    Ok(Some(db_user))
}

/// Get a user from the database by username or email
/// Returns a Result with the user if found, None if not found, or an error if there was a problem
#[server(endpoint = "find_user", client = Client)]
pub async fn get_user(username_or_email: String) -> BackendResult<Option<User>> {
    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let mut user = find_user(username_or_email, &mut db_conn)
        .await
        .context("Error finding user by username or email")?;

    // Remove the password hash before returning the user
    if let Some(user) = user.as_mut() {
        user.password = None;
    }

    Ok(user)
}

#[server(endpoint = "get_user_by_id", client = Client)]
pub async fn get_user_by_id(user_id: i32) -> BackendResult<Option<User>> {
    let mut db_conn = BackendState::get().await?.get_db_conn()?;

    let mut user = find_user_by_id(user_id, &mut db_conn)
        .await
        .context("Error finding user by ID")?;

    // Remove the password hash before returning the user
    if let Some(user) = user.as_mut() {
        user.password = None;
    }

    Ok(user)
}
