#![recursion_limit = "256"]

#[cfg(feature = "ssr")]
extern crate diesel_migrations;

#[cfg(feature = "ssr")]
#[tokio::main]
async fn main() {
    use axum::{body::Body, extract::Request, http::Response, middleware::Next};
    use axum::{extract::Path, middleware::from_fn, routing::get, Router};
    use axum_login::tower_sessions::SessionManagerLayer;
    use axum_login::AuthManagerLayerBuilder;
    use http::StatusCode;
    use leptos::prelude::*;
    use leptos_axum::{generate_route_list, LeptosRoutes};
    use libretunes::app::*;
    use libretunes::util::auth_backend::AuthBackend;
    use libretunes::util::backend_state::BackendState;
    use libretunes::util::config::load_config;
    use libretunes::util::fileserv::{
        file_and_error_handler, get_asset_file, get_static_file, AssetType,
    };
    use libretunes::util::require_auth::require_auth_middleware;
    use log::*;
    use tower_sessions_redis_store::RedisStore;

    flexi_logger::Logger::try_with_env_or_str("debug")
        .unwrap()
        .format(flexi_logger::opt_format)
        .start()
        .unwrap();

    info!("\n{}", include_str!("../ascii_art.txt"));

    let config = load_config().unwrap_or_else(|err| {
        error!("Failed to load configuration: {}", err);
        std::process::exit(1);
    });

    let state = BackendState::from_config(config)
        .await
        .unwrap_or_else(|err| {
            error!("Failed to initialize backend state: {}", err);
            std::process::exit(1);
        });

    info!("Starting Leptos server...");

    use dotenvy::dotenv;
    dotenv().ok();

    debug!("Running database migrations...");

    let mut db_conn = state.get_db_conn().unwrap_or_else(|err| {
        error!("Failed to get database connection: {}", err);
        std::process::exit(1);
    });

    // Bring the database up to date
    libretunes::util::database::migrate(&mut db_conn);
    drop(db_conn); // Close the connection after migrations

    debug!("Setting up session store...");
    let session_store = RedisStore::new(state.get_redis_conn());
    let session_layer = SessionManagerLayer::new(session_store);

    let auth_backend = AuthBackend {
        backend_state: state.clone(),
    };

    let auth_layer = AuthManagerLayerBuilder::new(auth_backend, session_layer).build();

    // A middleware that injects the backend state into the request extensions,
    // allowing it to be extracted later in the request lifecycle
    let backend_state_middleware = move |mut req: Request, next: Next| {
        let state = state.clone();

        async move {
            req.extensions_mut().insert(state);

            let response = next.run(req).await;
            Ok::<Response<Body>, (StatusCode, &'static str)>(response)
        }
    };

    let conf = get_configuration(None).unwrap();
    let leptos_options = conf.leptos_options;
    let addr = leptos_options.site_addr;
    // Generate the list of routes in your Leptos App
    let routes = generate_route_list(App);

    let app = Router::new()
        .leptos_routes(&leptos_options, routes, {
            let leptos_options = leptos_options.clone();
            move || shell(leptos_options.clone())
        })
        .route(
            "/assets/audio/{song}",
            get(|Path(song): Path<String>| get_asset_file(song, AssetType::Audio)),
        )
        .route(
            "/assets/images/{image}",
            get(|Path(image): Path<String>| get_asset_file(image, AssetType::Image)),
        )
        .route("/assets/{*uri}", get(|uri| get_static_file(uri, "")))
        .layer(from_fn(require_auth_middleware))
        .layer(auth_layer)
        .layer(from_fn(backend_state_middleware))
        .fallback(file_and_error_handler)
        .with_state(leptos_options);

    let listener = tokio::net::TcpListener::bind(&addr)
        .await
        .unwrap_or_else(|_| panic!("Could not bind to {}", &addr));

    info!("Listening on http://{}", &addr);

    axum::serve(listener, app.into_make_service())
        .await
        .expect("Server failed");
}

#[cfg(not(feature = "ssr"))]
pub fn main() {
    // no client-side main function
    // unless we want this to work with e.g., Trunk for pure client-side testing
    // see lib.rs for hydration function instead
}
