use crate::util::error::*;
use std::fs::File;
use symphonia::core::codecs::CodecType;
use symphonia::core::formats::FormatOptions;
use symphonia::core::io::MediaSourceStream;
use symphonia::core::meta::MetadataOptions;
use symphonia::core::probe::Hint;

/// Extract the codec and duration of an audio file
/// This is combined into one function because the file object will be consumed
pub fn extract_metadata(file: File) -> BackendResult<(CodecType, u64)> {
    let source_stream = MediaSourceStream::new(Box::new(file), Default::default());

    let hint = Hint::new();
    let format_opts = FormatOptions::default();
    let metadata_opts = MetadataOptions::default();

    let probe = symphonia::default::get_probe()
        .format(&hint, source_stream, &format_opts, &metadata_opts)
        .map_err(|e| InputError::InvalidInput(format!("{e}")))?;
    let reader = probe.format;

    if reader.tracks().len() != 1 {
        return Err(InputError::InvalidInput(format!(
            "Expected 1 track, found {}",
            reader.tracks().len()
        ))
        .into());
    }

    let track = &reader.tracks()[0];

    let time_base = track
        .codec_params
        .time_base
        .ok_or(InputError::InvalidInput("Missing time base".into()))?;
    let duration = track
        .codec_params
        .n_frames
        .map(|frames| track.codec_params.start_ts + frames)
        .ok_or(InputError::InvalidInput("Missing frame count".to_string()))?;

    let duration = duration
        .checked_mul(time_base.numer as u64)
        .and_then(|v| v.checked_div(time_base.denom as u64))
        .ok_or(BackendError::InternalError(
            "Duration calculation overflowed",
        ))?;

    Ok((track.codec_params.codec, duration))
}
