use crate::util::config::Config;
use crate::util::database::{PgPool, PgPooledConn};
use crate::util::error::*;
use axum::extract::FromRequestParts;
use diesel::{pg::PgConnection, r2d2::ConnectionManager};
use http::{request::Parts, StatusCode};
use leptos_axum::extract;
use std::ops::Deref;
use std::sync::Arc;
use tower_sessions_redis_store::fred::clients::Client as RedisClient;
use tower_sessions_redis_store::fred::clients::Pool as RedisPool;
use tower_sessions_redis_store::fred::prelude::*;
use tower_sessions_redis_store::fred::types::config::Config as RedisConfig;

/// Shared state of the backend. Contains the global server configuration,
/// a connection pool to the `PostgreSQL` database, and a connection pool to Redis.
#[derive(Clone)]
pub struct BackendState(Arc<BackendStateInner>);

impl Deref for BackendState {
    type Target = Arc<BackendStateInner>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl BackendState {
    /// Creates a new `BackendState` from the provided configuration
    pub async fn from_config(config: Config) -> BackendResult<Self> {
        let inner = BackendStateInner::from_config(config).await?;
        Ok(Self(Arc::new(inner)))
    }

    /// Extracts the `BackendState` from the request parts.
    pub async fn get() -> Result<Self, BackendError> {
        extract::<Self>().await.map_err(|e| {
            BackendError::InternalError(format!("Failed to extract BackendState: {e}"))
        })
    }
}

impl<S> FromRequestParts<S> for BackendState
where
    S: Send + Sync,
{
    // TODO(?) Change type to BackendError
    type Rejection = (StatusCode, &'static str);

    async fn from_request_parts(parts: &mut Parts, _state: &S) -> Result<Self, Self::Rejection> {
        parts
            .extensions
            .get::<BackendState>()
            .cloned()
            .ok_or((StatusCode::INTERNAL_SERVER_ERROR, "Backend state not found"))
    }
}

/// Inner state of the backend, containing the configuration and connection pools
pub struct BackendStateInner {
    pub config: Config,
    pub pg_pool: PgPool,
    pub redis_pool: RedisPool,
}

impl BackendStateInner {
    pub async fn from_config(config: Config) -> BackendResult<Self> {
        let pg_manager = ConnectionManager::<PgConnection>::new(config.database_url());

        let pg_pool = PgPool::builder().build(pg_manager).map_err(|e| {
            BackendError::InternalError(format!("Failed to create PostgreSQL pool: {e}"))
        })?;

        let redis_config = RedisConfig::from_url(&config.redis_url)
            .map_err(|e| BackendError::InternalError(format!("Failed to parse Redis URL: {e}")))?;

        let redis_pool = RedisPool::new(redis_config, None, None, None, 1).map_err(|e| {
            BackendError::InternalError(format!("Failed to create Redis pool: {e}"))
        })?;

        redis_pool.connect();

        redis_pool
            .wait_for_connect()
            .await
            .map_err(|e| BackendError::InternalError(format!("Failed to connect to Redis: {e}")))?;

        Ok(Self {
            config,
            pg_pool,
            redis_pool,
        })
    }

    pub fn get_db_conn(&self) -> BackendResult<PgPooledConn> {
        self.pg_pool.get().map_err(|e| {
            BackendError::InternalError(format!("Failed to get database connection: {e}"))
        })
    }

    pub fn get_redis_conn(&self) -> RedisClient {
        self.redis_pool.next().clone()
    }
}
