use leptos::prelude::*;
use serde::{Deserialize, Serialize};
use server_fn::codec::JsonEncoding;
use std::fmt::{Display, Formatter, Result as FmtResult};
use std::panic::Location;
use thiserror::Error;

/// A location in the source code
#[derive(Serialize, Deserialize, Debug, Clone)]
struct ErrorLocation {
    pub file: String,
    pub line: u32,
}

impl ErrorLocation {
    /// Creates a new `ErrorLocation` with the file and line number of the caller.
    #[track_caller]
    pub fn new() -> Self {
        let location = Location::caller();

        ErrorLocation {
            file: location.file().to_string(),
            line: location.line(),
        }
    }
}

impl Display for ErrorLocation {
    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
        write!(f, "{}:{}", self.file, self.line)
    }
}

pub type BackendResult<T> = Result<T, BackendError>;

/// A custom error type for backend errors
/// Contains the error, the location where it was created, and context added to the error.
#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct BackendError {
    /// The error type and data
    error: BackendErrorType,
    /// The location where the error was created
    from: ErrorLocation,
    /// Context added to the error, and location where it was added
    context: Vec<(ErrorLocation, String)>,
}

impl BackendError {
    /// Creates a new `BackendError` at this location with the given error type
    #[track_caller]
    fn new(error: BackendErrorType) -> Self {
        BackendError {
            error,
            from: ErrorLocation::new(),
            context: Vec::new(),
        }
    }

    /// Createa new `BackendError` from a `ServerFnErrorErr`
    #[track_caller]
    #[allow(non_snake_case)]
    pub fn ServerFnError(error: ServerFnErrorErr) -> Self {
        BackendError::new(BackendErrorType::ServerFn(error))
    }

    /// Creates a new `BackendError` from a Diesel error
    #[track_caller]
    #[allow(non_snake_case)]
    #[cfg(feature = "ssr")]
    pub fn DieselError(error: diesel::result::Error) -> Self {
        BackendError::new(BackendErrorType::Diesel(error.to_string()))
    }

    /// Creates a new `BackendError` from an I/O error
    #[track_caller]
    #[allow(non_snake_case)]
    pub fn IoError(error: std::io::Error) -> Self {
        BackendError::new(BackendErrorType::Io(error.to_string()))
    }

    /// Creates a new `BackendError` from an authentication error
    #[track_caller]
    #[allow(non_snake_case)]
    pub fn AuthError(error: AuthError) -> Self {
        BackendError::new(BackendErrorType::Auth(error))
    }

    /// Creates a new `BackendError` from an input error
    #[track_caller]
    #[allow(non_snake_case)]
    pub fn InputError(error: InputError) -> Self {
        BackendError::new(BackendErrorType::Input(error))
    }

    /// Creates a new `BackendError` from an access error
    #[track_caller]
    #[allow(non_snake_case)]
    pub fn AccessError(error: AccessError) -> Self {
        BackendError::new(BackendErrorType::Access(error))
    }

    /// Creates a new `BackendError` for a generic internal server error
    #[track_caller]
    #[allow(non_snake_case)]
    pub fn InternalError(message: impl Into<String>) -> Self {
        BackendError::new(BackendErrorType::Internal(message.into()))
    }

    /// Adds a context message to the error
    #[track_caller]
    #[allow(non_snake_case)]
    pub fn context(mut self, context: impl Into<String>) -> Self {
        self.context.push((ErrorLocation::new(), context.into()));
        self
    }

    /// Converts the error to a view component for display
    pub fn to_component(mut self) -> impl IntoView {
        use leptos_icons::*;

        leptos::logging::error!("{}", self);

        let error = self.error.to_string();
        self.context.reverse(); // Reverse the context to show the most recent first

        // Get the last context, if any, or the error message itself
        let message = self
            .context
            .first()
            .cloned()
            .map(|(_location, message)| message)
            .unwrap_or_else(|| error.clone());

        view! {
            <div class="text-red-800">
                <div class="grid grid-cols-[max-content_1fr] gap-1">
                    <Icon icon={icondata::BiErrorSolid} {..} class="self-center" />
                    <h1 class="self-center">{message}</h1>
                </div>
                {(!self.context.is_empty()).then(|| {
                    view! {
                        <details>
                            <summary class="cursor-pointer">{error.clone()}</summary>
                            <ul class="text-red-900">
                                {self.context.into_iter().map(|(location, message)| view! {
                                    <li>{format!("{location}: {message}")}</li>
                                }).collect::<Vec<_>>()}

                                <li>{format!("{}: {}", self.from, error)}</li>
                            </ul>
                        </details>
                    }
                })}
            </div>
        }
    }
}

impl Display for BackendError {
    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
        // Write the error type and its context
        write!(f, "BackendError: {}", self.error)?;

        if !self.context.is_empty() {
            write!(f, "\nContext:")?;

            for (location, message) in self.context.iter().rev() {
                write!(f, "\n  - {location}: {message}")?;
            }

            write!(f, "\n  - {}: {}", self.from, self.error)?;
        } else {
            write!(f, "\nFrom: {}", self.from)?;
        }

        Ok(())
    }
}

impl std::error::Error for BackendError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        Some(&self.error)
    }
}

impl<E: Into<BackendErrorType>> From<E> for BackendError {
    fn from(error: E) -> Self {
        BackendError::new(error.into())
    }
}

impl From<ServerFnErrorErr> for BackendError {
    fn from(error: ServerFnErrorErr) -> Self {
        BackendError::ServerFnError(error)
    }
}

pub trait Contextualize<R> {
    /// Add context to the `Result` if it is an `Err`.
    #[track_caller]
    fn context(self, context: impl Into<String>) -> R;
}

impl<T, E: Into<BackendError>> Contextualize<Result<T, BackendError>> for Result<T, E> {
    #[track_caller]
    fn context(self, context: impl Into<String>) -> Result<T, BackendError> {
        match self {
            Ok(value) => Ok(value),
            Err(e) => {
                // Convert the error into BackendError and add context
                let backend_error = e.into().context(context);
                Err(backend_error)
            }
        }
    }
}

impl<E: Into<BackendError>> Contextualize<BackendError> for E {
    #[track_caller]
    fn context(self, context: impl Into<String>) -> BackendError {
        self.into().context(context)
    }
}

/// The inner error type for `BackendError`
#[derive(Serialize, Deserialize, Debug, Clone, Error)]
enum BackendErrorType {
    #[error("Server function error: {0}")]
    ServerFn(ServerFnErrorErr),
    // Using string to represent Diesel errors,
    // because Diesel's Error type is not `Serializable`.
    #[error("Database error: {0}")]
    Diesel(String),

    /// Using `String` to represent I/O errors,
    /// because the `std::io::Error` type is not `Serializable`.
    #[error("I/O error: {0}")]
    Io(String),

    #[error("Authentication error: {0}")]
    Auth(AuthError),

    #[error("Input error: {0}")]
    Input(InputError),

    #[error("Access error: {0}")]
    Access(AccessError),

    #[error("Internal server error: {0}")]
    Internal(String),
}

impl FromServerFnError for BackendError {
    type Encoder = JsonEncoding;

    #[track_caller]
    fn from_server_fn_error(value: ServerFnErrorErr) -> Self {
        BackendError::new(BackendErrorType::ServerFn(value))
    }
}

#[cfg(feature = "ssr")]
impl From<diesel::result::Error> for BackendError {
    #[track_caller]
    fn from(err: diesel::result::Error) -> Self {
        BackendError::new(BackendErrorType::Diesel(format!("{err}")))
    }
}

impl From<std::io::Error> for BackendError {
    #[track_caller]
    fn from(err: std::io::Error) -> Self {
        BackendError::new(BackendErrorType::Io(format!("{err}")))
    }
}

impl From<AuthError> for BackendError {
    #[track_caller]
    fn from(err: AuthError) -> Self {
        BackendError::new(BackendErrorType::Auth(err))
    }
}

impl From<InputError> for BackendError {
    #[track_caller]
    fn from(err: InputError) -> Self {
        BackendError::new(BackendErrorType::Input(err))
    }
}

impl From<AccessError> for BackendError {
    #[track_caller]
    fn from(err: AccessError) -> Self {
        BackendError::new(BackendErrorType::Access(err))
    }
}

#[derive(Serialize, Deserialize, Debug, Clone, Error)]
pub enum AuthError {
    #[error("Unauthorized")]
    Unauthorized,
    #[error("Action requires admin privileges")]
    AdminRequired,
    #[error("Invalid credentials provided")]
    InvalidCredentials,
    #[error("Signup is disabled")]
    SignupDisabled,
    #[error("Error during authentication: {0}")]
    AuthError(String),
}

#[derive(Serialize, Deserialize, Debug, Clone, Error)]
pub enum InputError {
    #[error("Missing required field {0}")]
    MissingField(String),
    #[error("Error reading field: {0}")]
    FieldReadError(String),
    #[error("Invalid input: {0}")]
    InvalidInput(String),
}

#[derive(Serialize, Deserialize, Debug, Clone, Error)]
pub enum AccessError {
    #[error("Not found")]
    NotFound,
    #[error("Not found or unauthorized")]
    NotFoundOrUnauthorized,
}
