use cfg_if::cfg_if;

cfg_if! {
    if #[cfg(feature = "reqwest_api")] {
        pub type Client = CustomClient;

        use std::future::Future;
        use once_cell::sync::Lazy;
        use leptos::prelude::{FromServerFnError, ServerFnErrorErr};
        use server_fn::{Bytes, client::get_server_url, error::IntoAppError};
        use futures::{TryFutureExt, StreamExt, SinkExt};

        /// Static global client
        static CLIENT: Lazy<reqwest::Client> = Lazy::new(|| reqwest::Client::builder()
            .cookie_store(true)
            .build()
            .unwrap());

        /// A custom client for `server_fn` that uses `reqwest`.
        /// Very similar to `server_fn::client::reqwest::ReqwestClient`,
        /// but with the cookies enabled to allow for authentication.
        pub struct CustomClient;

        impl <
            Error: FromServerFnError,
            InputStreamError: FromServerFnError,
            OutputStreamError: FromServerFnError,
            > server_fn::client::Client<Error, InputStreamError, OutputStreamError> for CustomClient {
            type Request = reqwest::Request;
            type Response = reqwest::Response;

            fn send(
                req: Self::Request
            ) -> impl Future<Output = Result<Self::Response, Error>>
                 + Send {

                CLIENT.execute(req).map_err(|e| {
                        ServerFnErrorErr::Request(e.to_string()).into_app_error()
                })
            }

            async fn open_websocket(
                path: &str,
            ) -> Result<(
                    impl futures::Stream<Item = Result<Bytes, Bytes>> + Send + 'static,
                    impl futures::Sink<Bytes> + Send + 'static,
                ),
                Error,
            > {
                let mut websocket_server_url = get_server_url().to_string();

                if let Some(postfix) = websocket_server_url.strip_prefix("http://") {
                    websocket_server_url = format!("ws://{postfix}");
                } else if let Some(postfix) = websocket_server_url.strip_prefix("https://") {
                    websocket_server_url = format!("wss://{postfix}");
                }

                let url = format!("{websocket_server_url}{path}");

                let (ws_stream, _) =
                    tokio_tungstenite::connect_async(url).await.map_err(|e| {
                        Error::from_server_fn_error(ServerFnErrorErr::Request(
                            e.to_string(),
                        ))
                    })?;

                let (write, read) = ws_stream.split();

                Ok((
                    read.map(|msg| match msg {
                        Ok(msg) => Ok(msg.into_data()),
                        Err(e) => Err(OutputStreamError::from_server_fn_error(
                            ServerFnErrorErr::Request(e.to_string()),
                        )
                        .ser()),
                    }),

                    write.with(|msg: Bytes| async move {
                        Ok::<
                            tokio_tungstenite::tungstenite::Message,
                            tokio_tungstenite::tungstenite::Error,
                        >(
                            tokio_tungstenite::tungstenite::Message::Binary(msg)
                        )
                    }),
                ))
            }

            fn spawn(future: impl Future<Output = ()> + Send + 'static) {
                tokio::spawn(future);
            }
        }
    } else {
        pub type Client = server_fn::client::browser::BrowserClient;
    }
}
