use leptos::logging::*;
use leptos::prelude::*;

use crate::api::auth::get_logged_in_user;
use crate::api::playlists::get_playlists;
use crate::models::backend;
use crate::models::backend::User;
use crate::models::frontend::PlayStatus;
use crate::util::error::*;

/// Global front-end state
/// Contains anything frequently needed across multiple components
/// Behaves like a singleton, in that `provide_context`/`expect_context` will
/// always return the same instance
#[derive(Clone)]
pub struct GlobalState {
    /// A resource that fetches the logged in user
    /// This will not automatically refetch, so any login/logout related code
    /// should call `refetch` on this resource
    pub logged_in_user: Resource<Option<User>>,

    /// The current play status
    pub play_status: RwSignal<PlayStatus>,

    /// A resource that fetches the playlists
    pub playlists: Resource<BackendResult<Vec<backend::Playlist>>>,
}

impl GlobalState {
    pub fn new() -> Self {
        let play_status = RwSignal::new(PlayStatus::default());

        let logged_in_user = Resource::new(
            || (),
            |_| async {
                get_logged_in_user()
                    .await
                    .inspect_err(|e| {
                        error!("Error getting logged in user: {:?}", e);
                    })
                    .ok()
                    .flatten()
            },
        );

        // Refetch playlists when the logged in user changes
        let playlists = Resource::new(move || logged_in_user.track(), |_| get_playlists());

        Self {
            logged_in_user,
            play_status,
            playlists,
        }
    }

    pub fn logged_in_user() -> Resource<Option<User>> {
        expect_context::<Self>().logged_in_user
    }

    pub fn play_status() -> RwSignal<PlayStatus> {
        expect_context::<Self>().play_status
    }

    pub fn playlists() -> Resource<BackendResult<Vec<backend::Playlist>>> {
        expect_context::<Self>().playlists
    }
}

impl Default for GlobalState {
    fn default() -> Self {
        Self::new()
    }
}
